local core = CalendarExtras

local module = core:NewModule("Status Online")

 --[[ -------- Locals ------- ]]--

local listIsDirty

local assumeMissingAreOnline -- If true, treat players not found in guild/friend list as being online

local sortersModule

local listFriends = {}
local listGuild = {}
local altLookup
--local listChannels = setmetatable({}, meta) -- Need to figure out how to do this easily...

local libAlts = LibStub("LibAlts-1.0")

local timer = CreateFrame("frame") -- Minimal timer to minimally cut down like one whole list update call, maybe. Like wow

local classNamesReverseLookup = {}
for k, v in pairs(LOCALIZED_CLASS_NAMES_MALE) do
	classNamesReverseLookup[v] = k
end
for k, v in pairs(LOCALIZED_CLASS_NAMES_FEMALE) do
	classNamesReverseLookup[v] = k
end

 --[[ -------- Timer -------- ]]--

do
	local totalElapsed = 0
	timer.Start, timer.Stop = timer.Show, timer.Hide -- Useless fluff, I know
	timer:SetScript("OnUpdate", function(self, elapsed)
		totalElapsed = totalElapsed + elapsed
		if totalElapsed >= 2 then
			totalElapsed = 0
			self:Stop()
			if listIsDirty then
				listIsDirty = nil
				core:UpdateOpenList() -- This is a bit heavy call I guess, but I don't want to recolorize all the entries myself
			end
		end
	end)
end

 --[[ ----------------------- ]]--

local function Capitalize(str)
	if not str then return end
	return str:sub(1, 1):upper()..str:sub(2)
end

local Colorize
do
	local classColorStr = "|cff%.2x%.2x%.2x%s|r"
	function Colorzie(str, class)
		if not str or not class then return end
		local c = RAID_CLASS_COLORS[class]
		if not c then return str end
		return classColorStr:format(c.r*255, c.g*255, c.b*255, str)
	end
end

local DimButton
do
	local buttons = {}
	function DimButton(button)
			local buttonName = button:GetName()
			local name = _G[buttonName.."Name"]
			buttons[1], buttons[2], buttons[3] = _G[buttonName.."Class"], _G[buttonName.."Status"], sortersModule and _G[buttonName.."Role"] or nil
			name:SetTextColor(0.6, 0.6, 0.6, 0.65)
			local r, g, b
			for _, button in ipairs(buttons) do
				r, g, b = button:GetTextColor()
				button:SetTextColor(r, g, b, 0.5)
			end
	end
end

-- Returns the person's class if he's online, otherwise nil. This will get very confusing very quick >_>
local function IsOnline(name)
	return listGuild[name] or listFriends[name] or (UnitInRaid(name) and UnitIsConnected(name)) or (UnitInParty(name) and UnitIsConnected(name)) --[[ or listChannels[name] ]]
end

-- Returns either boolean if the name doesn't have any main-alt relationships, otherwise it returns the class-colored alt name string. This is a total clusterfuck.
local function ProcessPlayer(name)
	local class = IsOnline(name)
	if class then
		--print(name, "is online on main")
		return true
	else -- I kinda dislike this, it seems like a lot of stuff to do on each offline player in the list, each time the list update fires (which it does pretty often too), but can't figure out a more elegant way right now
		local main = libAlts:IsMain(name) and name or libAlts:GetMain(name)
		if main and main ~= name then
			--print(name, "has a main called", main)
			class = IsOnline(main)
			if class then
				--print(name, "'s main is online")
				return Colorize(main, class)
			end
		end
		if not online and main then
			local t = libAlts.Alts[main:lower()] -- This is probably not a good idea, but why is the LibAlts:GetAlts() unpacking the return value for me? That's no good O_o
			if t and #t > 0 then
				for _, alt in pairs(t) do
					if alt ~= name then
						alt = Capitalize(alt)
						class = IsOnline(alt)
						--print("Checking", name, "for alt:", alt)
						if class then
							--print(alt, "is online, class:", class)
							return Colorzie(alt, class)
						end
					end
				end
			end
		end
	end
end

function module:FRIENDLIST_UPDATE()
	for i = 1, GetNumFriends() do
		local name, _, class, _, online = GetFriendInfo(i)
		local status
		if (assumeMissingAreOnline and not online) or (not assumeMissingAreOnline and online) then
			status = classNamesReverseLookup[class]
		end
		if not listIsDirty and listFriends[name] ~= status then
			listIsDirty = true
			altLookup = nil
			timer:Start()
		end
		listFriends[name] = status
	end
end

function module:GUILD_ROSTER_UPDATE()
	for i = 1, GetNumGuildMembers(true) do
		local name, _, _, _, _, _, _, _, online, _, class = GetGuildRosterInfo(i)
		local status
		if (assumeMissingAreOnline and not online) or (not assumeMissingAreOnline and online) then
			status = class
		end
		if not listIsDirty and listGuild[name] ~= status then
			listIsDirty = true
			altLookup = nil
			timer:Start()
		end
		listGuild[name] = status
	end
end

-- If a person is online but on an alt, note this in the tooltip when mouseovering the entry in the list
local function OnEnter(frame)
	if not altLookup then return end
	local name = CalendarEventGetInvite(frame.inviteIndex)
	local alt = altLookup[name]
	if alt then
		if not GameTooltip:GetOwner() then
			GameTooltip:SetOwner(frame, "ANCHOR_TOPRIGHT")
		else
			GameTooltip:AddLine(" ")
		end
		GameTooltip:AddLine("(On his/her alt "..alt, 0.6, 0.6, 0.6)
		GameTooltip:Show()
	end
end

-- Dim buttons for the offline people. This fires after the original Calendar***EventInviteListScrollFrame_Update which changes the colors for the buttons, so no need to "undim" them
local function UpdateInviteListButtons(_, frame)
	for _, button in ipairs(frame.buttons) do
		local inviteIndex = button.inviteIndex
		if inviteIndex then
			local name = CalendarEventGetInvite(inviteIndex)
			local found = ProcessPlayer(name)
			if type(found) == "string" then
				if not altLookup then altLookup = {} end
				altLookup[name] = found -- Remember which alt this person is logged on on. I suck at naming variables :-(
				if not button:GetScript("OnEnter") then
					button:SetScript("OnEnter", OnEnter)
				end
			end
			if (assumeMissingAreOnline and found) or (not assumeMissingAreOnline and not found) then
				DimButton(button)
			end
		end
	end
end

function module:OnInitialize()
	if core.modules["Sorters"] then
		sortersModule = core.modules["Sorters"]
	end

	core.RegisterCallback(self, "Post_Update", UpdateInviteListButtons)

	-- Need to refresh things if people go online/offline, hopefully these catch all things properly.
	core.RegisterCallback(self, "CalendarShow", function()
		self:RegisterEvent("GUILD_ROSTER_UPDATE")
		self:RegisterEvent("FRIENDLIST_UPDATE")
		self:GUILD_ROSTER_UPDATE()
		self:FRIENDLIST_UPDATE()
	end)       
	core.RegisterCallback(self, "CalendarHide", function()
		self:UnregisterEvent("GUILD_ROSTER_UPDATE")
		self:UnregisterEvent("FRIENDLIST_UPDATE")
		timer:Stop()
	end)
	self:RegisterEvent("GUILD_ROSTER_UPDATE")
	self:RegisterEvent("FRIENDLIST_UPDATE")
	self:GUILD_ROSTER_UPDATE()
	self:FRIENDLIST_UPDATE()
	GuildRoster()

	-- Post-hook to add alt-info to the existing tooltips
	self:RawHook("CalendarEventInviteListButton_OnEnter", function(frame, ...)
		self.hooks.CalendarEventInviteListButton_OnEnter(frame, ...)
		OnEnter(frame)
	end, true)

	-- Add whisper-to-alt for easy yelling at n00bs who're still on an alt at raid start! >:-(
	core:NewContextMenuEntry({
		text = "Whisper alt",
		func = self.ContextMenuFunc,
		enable = self.ContextMenuEnableCheck,
	})

	local function f()
		listIsDirty = true
		timer:Start()
	end
	libAlts.RegisterCallback(self, "LibAlts_SetAlt", f)
	libAlts.RegisterCallback(self, "LibAlts_RemoveAlt", f)
end

 --[[ ----- Context menu ---- ]]--

-- Only enable for the players that are on an alt
function module.ContextMenuEnableCheck(inviteIndex)
	return altLookup and altLookup[(CalendarEventGetInvite(inviteIndex))]
end

-- Open whisper to the alt, and then let 'em have it! This'll be the last time that slacker misses a raid invite...
function module.ContextMenuFunc()
	if not altLookup then return end
	local name = CalendarEventGetInvite(core.ContextMenuGetInviteButton().inviteIndex)
	local alt = altLookup[name]:sub(11, -3)
	if alt then
		ChatFrame_SendTell(alt)
	end
end

--[[
	Some of the channel functions I found
local channel, header, collapsed, channelNumber, count, active, category = GetChannelDisplayInfo(id)
local name, owner, moderator, muted, active, enabled = GetChannelRosterInfo(id, rosterIndex)
SetSelectedDisplayChannel(id)
CHANNEL_ROSTER_UPDATE(channelId)
]]